# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from typing import Any

import pelican
from pelican import signals

from . import PLUMAGE_ROOT
from .dom_transforms import transform
from .favicon import add_favicon_assets
from .webassets import setup_webassets

ALL_CODE_STYLES: set[str] = {
    f.stem for f in PLUMAGE_ROOT.joinpath("static/css/pygments/").resolve().iterdir()
}


def register_signals() -> None:
    signals.initialized.connect(check_config)
    signals.static_generator_finalized.connect(add_favicon_assets)
    signals.content_written.connect(transform)


def check_config(sender) -> None:
    """Validates and setup Plumage configuration."""
    conf = sender.settings

    # Keep some metadata around.
    conf["PELICAN_VERSION"] = pelican.__version__

    # Check code coloe scheme ID.
    conf = check_codestyle(conf)

    # Setup webassets plugin.
    conf = setup_webassets(conf)

    sender.settings = conf


def check_codestyle(conf: dict[str, Any]) -> dict[str, Any]:
    """Check selected code color scheme ID is recognized by Pygments.

    Defaults to "monokai" if none set in Pelican's conf.
    """
    # Defaults code style to Monokai.
    if not conf.get("CODE_STYLE"):
        conf["CODE_STYLE"] = "monokai"

    code_style = conf["CODE_STYLE"]
    if code_style not in ALL_CODE_STYLES:
        msg = f"{code_style} not recognized among {sorted(ALL_CODE_STYLES)}."
        raise ValueError(
            msg,
        )

    return conf



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import logging
from pathlib import Path

__version__ = "4.0.0"
""" Examples of valid version strings according :pep:`440#version-scheme`:

.. code-block:: python
    __version__ = "1.2.3.dev1"  # Development release 1
    __version__ = "1.2.3a1"  # Alpha Release 1
    __version__ = "1.2.3b1"  # Beta Release 1
    __version__ = "1.2.3rc1"  # RC Release 1
    __version__ = "1.2.3"  # Final Release
    __version__ = "1.2.3.post1"  # Post Release 1
"""


logger = logging.getLogger(__name__)

PLUMAGE_ROOT: Path = Path(__file__).resolve().parent


def get_path() -> str:
    """Returns installation path of the theme.

    Used in ``pelicanconf.py`` to dynamiccaly fetch theme location on the system.
    """
    return str(PLUMAGE_ROOT)


# Workaround circular dependencies.
from .config import register_signals  # noqa: E402

register_signals()



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from urllib.parse import urlparse

from pyquery import PyQuery as pq


def transform(path, context):
    BASE_URL = urlparse(context.get("SITEURL", ""))

    # XXX This direct construct is stripping the "<!DOCTYPE>" heading. See:
    # https://github.com/gawel/pyquery/issues/199
    doc = pq(filename=path)

    # Add bootstrap table style to table elements.
    doc("#content table").add_class("table table-hover")
    doc("#content table thead th").attr("scope", "col")

    # Make images responsive and styled in article content, but ignore
    # images in cards (like those from project template), and images attached to
    # links.
    main_images_selector = "#content img:not(.card-img-top):not(.link-icon)"
    doc(main_images_selector).add_class("img-fluid border rounded shadow")

    # Process all images from the main content to create a reduced set with
    # lower dimensions.
    # XXX Hack to bypass the bug on external images from image-process
    # plugin: https://github.com/pelican-plugins/image-process/issues/33

    def exclude_external_images(_, this):
        source = urlparse(this.get("src", ""))
        if (source.scheme and BASE_URL.scheme) and (source.netloc != BASE_URL.netloc):
            return False
        return True

    doc(main_images_selector).filter(exclude_external_images).add_class(
        "image-process-article-photo",
    )

    # Style blockquotes in the way Bootstrap does.
    doc("blockquote").add_class(
        "blockquote border-start border-primary-subtle bg-dark-subtle fs-6 border-4 ps-2",
    )
    doc("blockquote p").add_class("p-2")

    # Style code boxes.
    doc(".codehilite, .highlight").add_class(
        f"pygments-style-{context.get('CODE_STYLE')} rounded shadow-sm mb-3",
    )

    # Style admonitions produced by Python Markdown into alerts.
    doc(".admonition").add_class("alert shadow").attr("role", "alert")
    doc(".admonition-title").add_class("alert-heading h4")
    # Map rST's admonition types to Bootstrap's:
    # https://python-markdown.github.io/extensions/admonition/#syntax
    admo_map = {
        "primary": {"primary"},
        "secondary": {"secondary"},
        "success": {"success"},
        "danger": {"danger", "error"},
        "warning": {"warning", "attention", "caution", "important"},
        "info": {"info", "hint", "note", "tip"},
        "light": {"light"},
        "dark": {"dark"},
    }
    for bootstrap_class, admo_classes in admo_map.items():
        for admo_class in admo_classes:
            doc(f".admonition.{admo_class}").add_class(f"alert-{bootstrap_class}")

    # Save result.
    with open(path, "w") as source:
        source.write(doc.outerHtml())



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


from pelican import signals
from pelican.contents import Static

from . import PLUMAGE_ROOT


def add_favicon_assets(sender):
    """Copy all individual files found in /static/favicon theme's folder to the root of
    the generated site.

    Favicons were generated with RealFaviconGenerator v0.16:
    https://realfavicongenerator.net
    """
    # Our theme structure expect default configuration.
    # See: https://docs.getpelican.com/en/latest/settings.html#themes
    assert "static" in sender.settings["THEME_STATIC_PATHS"]

    for asset in (PLUMAGE_ROOT / "static" / "favicon").iterdir():
        assert asset.is_file()

        # Add static file to generator context, as Pelican does natively at:
        # https://github.com/getpelican/pelican/blob/8033162ba4393db60791b201fb100d1be0f04431/pelican/generators.py#L811-L817
        static = sender.readers.read_file(
            base_path=asset.parent,
            path=asset.name,
            content_class=Static,
            fmt="static",
            context=sender.context,
            preread_signal=signals.static_generator_preread,
            preread_sender=sender,
            context_signal=signals.static_generator_context,
            context_sender=sender,
        )

        # Forces the asset to be saved at the root of the output folder.
        # See:
        # https://github.com/getpelican/pelican/blob/8033162ba4393db60791b201fb100d1be0f04431/pelican/contents.py#L55-L59
        static.metadata["save_as"] = asset.name
        static.override_save_as = asset.name

        # Register the new asset along other static files.
        sender.context["staticfiles"].append(static)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import importlib
import os
from pathlib import Path
from shutil import which
from textwrap import indent
from typing import Any

from pynpm import NPMPackage

from . import PLUMAGE_ROOT, logger

"""Setup the `webassets <https://github.com/miracle2k/webassets>`_ plugin for Pelican.
"""


CONFIG_DEFAULTS: dict[str, str] = {
    # No need to force a compressed rendering, a minification pass is applied
    # at the end of the pipeline. See:
    # https://webassets.readthedocs.io/en/latest/builtin_filters.html#webassets.filter.libsass.LibSass
    # "LIBSASS_STYLE": "compressed",
}
"""Default configuration for `webassets <https://github.com/miracle2k/webassets>`_.

See the `list of configuration parameters for each filter
<https://webassets.readthedocs.io/en/latest/builtin_filters.html>`_.
"""


POSTCSS_CLI_NAME = "postcss"
"""Name of the PostCSS CLI binary."""

CLOSURE_JAR_NAME = "closure.jar"
"""Name of the Closure CLI jar file."""


def postcss_config():
    """Produce the default configuration for PostCSS.

    Takes care of:

        - Installing PostCSS CLI and its dependency with ``npm``.
        - Locating the PostCSS CLI binary on the filesystem.
        - Registering PostCSS to the webassets plugin.
    """
    # The dependency definition file relative to Plumage's install path takes
    # precedence.
    node_deps_file = PLUMAGE_ROOT.joinpath("package.json").resolve()
    node_bin_path = node_deps_file.parent / "node_modules" / ".bin"
    cli_search_path = [
        str(node_bin_path),
    ]

    # Check if the path exist in any of the environment locations.
    env_path = os.pathsep.join([*cli_search_path, os.getenv("PATH")])
    postcss_bin = which(POSTCSS_CLI_NAME, path=env_path)

    if not postcss_bin:
        logger.warning(f"{POSTCSS_CLI_NAME} CLI not found.")

        # Install Node dependencies.
        logger.info(
            f"Install Plumage's Node.js dependencies from {node_deps_file}:\n"
            f"{indent(node_deps_file.read_text(), ' ' * 2)}",
        )
        pkg = NPMPackage(node_deps_file)
        try:
            pkg.install()
        except FileNotFoundError:
            logger.error("npm CLI not found.")
            raise

        postcss_bin = which(POSTCSS_CLI_NAME, path=env_path)
        assert postcss_bin

    logger.info(f"{POSTCSS_CLI_NAME} CLI found at {postcss_bin}")

    return {
        "POSTCSS_BIN": postcss_bin,
        # Force usage of autoprefixer via PostCSS.
        "POSTCSS_EXTRA_ARGS": ["--use", "autoprefixer"],
    }


def closure_config():
    """Locate the Closure CLI jar file provided by the ``closure`` package.

    .. warning::

        We need to locate the ``closure`` Python package without loading it, as it is
        old and unmaintained and produce the following error on calling
        ``import closure``:

            .. code-block:: pytb
                Traceback (most recent call last):
                File "<string>", line 1, in <module>
                File ".../python3.12/site-packages/closure/__init__.py", line 3, in <module>
                    from pkg_resources import resource_filename
                ModuleNotFoundError: No module named 'pkg_resources'
    """
    package = importlib.util.find_spec("closure")
    closure_jar = (Path(package.origin).parent / CLOSURE_JAR_NAME).resolve()
    logger.info(f"{CLOSURE_JAR_NAME} JAR file found at {closure_jar}")
    return {
        "CLOSURE_COMPRESSOR_PATH": str(closure_jar),
    }


def setup_webassets(conf: dict[str, Any]) -> dict[str, Any]:
    """Setup pelican-webassets plugin configuration."""
    # Merge static and dynamic configurations to produce webassets' defaults.
    default_conf = CONFIG_DEFAULTS.copy() | postcss_config() | closure_config()

    # Update the default configuration with user-defined values.
    webassets_conf = default_conf | dict(conf.get("WEBASSETS_CONFIG", {}))

    # Save updated configuration in Pelican settings in the form of ``(key, value)``
    # instead of a ``dict`` as expected by ``webassets`` plugin. See:
    # https://github.com/pelican-plugins/webassets/blob/2.0.0/README.md#webassets_config
    conf["WEBASSETS_CONFIG"] = list(webassets_conf.items())

    return conf
